function [SD_VLAD] = SD_VLAD_encoding(desc, vocab, st_d, doInnerProduct)
% [SD_VLAD] = SD_VLAD_encoding(desc, vocab, st_d)
%Compute SD_VLAD encoding for a set of descriptors
% SD-VLAD = Shape Difference for VLAD
%
% input:
% desc:           N x D row-wise matrix of descriptors
%                 (N - number of descriptors,
%                 D - the dimensionality of descriptors)

% vocab:          K x D row-wise visual vocabulary
%                 (K - number of visual words (centers),
%                 D the dimensionality of each visual word)

% st_d:           K x D row-wise standard deviations coresponding to 
%                 each visual work learned
% doInnerProduct: Boolean (0 1). Default: true(1), which means use inner 
%                 product to determine visual word assignment. Only   
%                 works when both descriptors and vocabulary are 
%                 normalized to unit length.                  
%                 If false(0): use Euclidean distance for the assignment
%                 step
%  output
%    SD_VLAD: 1 x 2*(K * D) SD_VLAD encoding vector.
%
%       Ionut Cosmin Duta - 2016


if nargin == 3
    doInnerProduct = 1;
end


dimDesc=size(desc, 2); % the dimensionality of the descriptors
K=size(vocab, 1);  % the size of the vocabulary

if doInnerProduct
    % Calculate similarity as inner product (arccos angle)
    %take care that desc and vocab are  normalized to unit length !!!           
    theSim = desc * vocab';
    [~, assign] = max(theSim, [], 2);
else
    % Calculate similarity using Euclidean distance
    theDist = distmj(desc, vocab);
    [~, assign] = min(theDist, [], 2);
end



word_SD_VLAD1=cell(1, K); %for first order statistics
word_SD_VLAD2=cell(1, K); %for standard deviation diferences

%Calculate SD_VLAD for each visual word of the vocabulary
for i=1:K
    
    assigned=(assign==i); % get the descriptors assigned to the cluster i;
    nAssigned=sum(assigned);
    if nAssigned>0 % compute SD_VLAD for each visual word (cluster) that has at least one descriptor assigned
        

        %compute the difference between descriptors and visual word (cluster)
        diff=bsxfun(@minus, desc(assigned, :), vocab(i, :));
        
        
        %Calculate the sum over these differences       
        %compute SD_VLAD1
        word_SD_VLAD1{i}=(1.0/nAssigned)*(sum(diff, 1) ./ st_d(i, :));
        
        %compute SD_VLAD2
        word_SD_VLAD2{i} = (((1.0/nAssigned)*(sum(diff.^2, 1))).^(1/2) - (st_d(i, :)));
        
            
        %check if NaN is resulted from the division with standard deviation 
        if sum(isnan(word_SD_VLAD1{i}))>0   
             word_SD_VLAD1{i}(isnan(word_SD_VLAD1{i}))=0;
        end
        
        %apply intra-normalization L2 for each component
        %comment these two lines if you don't want to use intra-normalization
        word_SD_VLAD1{i}=NormalizeRowsUnit(word_SD_VLAD1{i});
        word_SD_VLAD2{i}=NormalizeRowsUnit(word_SD_VLAD2{i});
        
   
    % no desccriptors are assigned to the cluster then put zeros
    else
        word_SD_VLAD1{i}=zeros(1, dimDesc);    
        word_SD_VLAD2{i}=zeros(1, dimDesc);   
    end 
                
end

%Concatenate all the VLAD vectors for each cluster to create the final SD-VLAD vector
word_SD_VLAD1=cat(2,word_SD_VLAD1{:});
word_SD_VLAD2=cat(2,word_SD_VLAD2{:});

SD_VLAD=cat(2, word_SD_VLAD1, word_SD_VLAD2);
